@@ -42,6 +42,9 @@
 #include <gnome.h>
 #include <gdk/gdkkeysyms.h>
 
+#include <libgnorba/gnorba.h>
+#include <libgnomeui/gnome-window-icon.h>
+
 #include "linux-cdrom.h"
 
 #include "gtracked.h"
@@ -50,6 +53,7 @@
 #include "callbacks.h"
 #include "gtcd_public.h"
 #include "keybindings.h"
+#include "gtcd.h"
 
 /* time display types */
 enum { TIME_FIRST=-1, TRACK_E, TRACK_R, DISC_E, DISC_R, TIME_LAST };
@@ -90,6 +94,7 @@ GdkFont *sfont, *tfont;
 GdkColor track_color, darkgrey, timecolor, blue;
 GdkGC *gc;
 GtkAccelGroup *accel;
+guint slow_timeout=0, fast_timeout=0;
 
 GtkTooltips *tooltips;
 
@@ -112,17 +117,20 @@ gint slow_timer(gpointer *data);
 gint roll_timer(gpointer *data);
 gint release_timer(gpointer *data);
 int skip_cb(GtkWidget *widget, GdkEvent *event, gpointer *data);
+int key_skip_cb(GtkWidget *widget, gpointer *data);
 gint changer_callback(GtkWidget *widget, gpointer *data);
 GtkWidget* make_changer_buttons(void);
 GtkWidget* create_buttons(void);
 void setup_colors(void);
 void draw_time_playing(void);
+void calculate_title(int, int);
 void draw_titles(void);
 void draw_time_scanning(void);
 gint volume_changed(GtkWidget *widget, gpointer *data);
 gint launch_gmix(GtkWidget *widget, GdkEvent *event, gpointer data);
 gint fast_timer(gpointer *data);
 void setup_time_display(GtkWidget *table);
+void calculate_title(int maxwidth, int inc);
 void setup_fonts(void);
 void init_window(void);
 GtkWidget* make_button_with_pixmap(char *pic, GtkSignalFunc func,
@@ -162,6 +170,25 @@ gint release_timer(gpointer *data)
 	return FALSE;
 }
 
+/*
+ * The callback routine for the visible next/prev (ff/rewind)
+ * buttons. Upon a BUTTON_PRESS event a timer is started and this
+ * routine exits.  If a BUTTON_RELEASE event is received before the
+ * timer expires, then skip forward/backward to the next track.  This
+ * routine performs the changes necessary for track skips.  If a
+ * BUTTON_RELEASE event is not received before the timer expires, then
+ * this is a fast forward/rewind operation.  Most of the work has
+ * already been performed inside of one of the timer functions.
+ *
+ * @param widget Unused.
+ *
+ * @param event The GDK event that occurred.  The interesting events
+ * for this routine are GDK_BUTTON_PRESS and GDK_BUTTON_RELEASE.  This
+ * routine also sees GDK_ENTER_NOTIFY and GDK_LEAVE_NOTIFY events.
+ *
+ * @param data An indication of whether to move forwards (1) or
+ * backwards (-1).
+ */
 int skip_cb(GtkWidget *widget, GdkEvent *event, gpointer *data)
 {
 	skip_first = TRUE;
@@ -178,8 +205,9 @@ int skip_cb(GtkWidget *widget, GdkEvent *event, gpointer *data)
 			gtk_timeout_remove(release_t);
 			if(GPOINTER_TO_INT(data) > 0)
 			{
-				if(cd.cur_t < cd.last_t)
-				{   
+				if((cd.cur_t < cd.last_t) &&
+				   (cd.trk[cd.cur_t+1].toc.cdte_ctrl != CDROM_DATA_TRACK))
+				{
 					cd.cur_t++;
 					tcd_playtracks(&cd,cd.cur_t, cd.last_t, prefs->only_use_trkind);
 					if(cd.play_method==REPEAT_TRK)
@@ -212,6 +240,61 @@ int skip_cb(GtkWidget *widget, GdkEvent *event, gpointer *data)
 	return FALSE;	
 }
 
+/*
+ * The callback routine for the accelerator keys on all of the
+ * invisible buttons.  These include the next/prev accelerators, fast
+ * forward/rewind accelerators, and volume up/down accelerators.  This
+ * routine is called once each time an appropriate accelerator key is
+ * pressed.
+ *
+ * @param widget Unused.
+ *
+ * @param data An indication of which accelerator key was pressed.
+ */
+int key_skip_cb(GtkWidget *widget, gpointer *data)
+{
+	gfloat value;
+
+	switch (GPOINTER_TO_INT(data)) {
+	    case 'F':
+		if(cd.cur_t < cd.last_t)
+		{   
+			cd.cur_t++;
+			tcd_playtracks(&cd,cd.cur_t, cd.last_t, prefs->only_use_trkind);
+			if(cd.play_method==REPEAT_TRK)
+				cd.repeat_track = cd.cur_t;
+		}
+		break;
+	    case 'B':
+		if( cd.cur_t > cd.first_t )
+		{
+			if( (cd.t_sec+(cd.t_min*60)) < 10 )
+				cd.cur_t--;
+			tcd_playtracks( &cd,cd.cur_t, cd.last_t, prefs->only_use_trkind);
+			if( cd.play_method==REPEAT_TRK )
+				cd.repeat_track = cd.cur_t;
+		}
+		break;
+	    case '[':
+		tcd_play_seconds(&cd, -4);
+		cd.repeat_track = cd.cur_t;
+		break;
+	    case ']':
+		tcd_play_seconds(&cd, 4);
+		cd.repeat_track = cd.cur_t;
+		break;
+	    case '<':
+	        value = GTK_ADJUSTMENT(vol)->value - 1;
+		gtk_adjustment_set_value(GTK_ADJUSTMENT(vol), value);
+		break;
+	    case '>':
+	        value = GTK_ADJUSTMENT(vol)->value + 1;
+		gtk_adjustment_set_value(GTK_ADJUSTMENT(vol), value);
+		break;
+	}
+	return FALSE;
+}   
+
 GtkWidget* make_button_with_pixmap(char *pic, GtkSignalFunc func,
 				   Shortcut *key,  gchar *tooltip)
 {
@@ -310,6 +393,51 @@ static gint button_press (GtkWidget *widget, GdkEvent *event)
 	return FALSE;
 }
 
+/*
+ * Create a button widget and add it to a table, but not as a visible
+ * widget.  Also call the routine to set of the accelerator binding
+ * for this button.
+ *
+ * @param table The table currently being built.
+ *
+ * @param label A user visible label that will appear in the
+ * preferences dialog for changing key mappings.
+ *
+ * @param obj_name The name used to add this button to the table.  The
+ * button is added as user data, not inserted into a row/column of the
+ * table.
+ *
+ * @param id The identifier used to tell the key_skip_cb routine what
+ * function to perform.
+ *
+ * @param key The accelerator key (and modifiers) to set.
+ */
+static void add_invisible(GtkWidget *table, gchar *label,
+			  gchar *obj_name, gint id, Shortcut *key)
+{
+	GtkWidget *button;
+
+	button = gtk_button_new();
+	gtk_signal_connect(GTK_OBJECT(button), "clicked",
+			   GTK_SIGNAL_FUNC(key_skip_cb), GINT_TO_POINTER(id));
+	gtk_object_set_data(GTK_OBJECT(table), obj_name, button);
+	add_key_binding(button, "clicked", label, key);
+}
+
+/*
+ * Create the table of CD Player controls.  This routine creates a 3x3
+ * matrix of visible buttons.  It also creates some invisible buttons
+ * to support the various accelerator keys.  These invisible buttons
+ * are required because a button press and release on the visible
+ * ff/rewind buttons (for example) is the equivalent of clicking that
+ * button.  Putting the accelerators onto the visible buttons and
+ * using the "clicked" signal results in all ff/rewinds becoming track
+ * skips, and all track skips being doubled.  These invisible buttons
+ * can be removed if there is a better method of implementing this, or
+ * if some other signal can be used.
+ *
+ * @return GtkWidget* A pointer to the newly created table
+ */
 GtkWidget* create_buttons(void)
 {
 	GtkWidget *table;
@@ -341,15 +469,12 @@ GtkWidget* create_buttons(void)
 	gtk_widget_set_events(ff, GDK_BUTTON_PRESS_MASK
 			      | GDK_BUTTON_RELEASE_MASK);
 
-	add_key_binding(rw, "clicked", _("Skip backwards"), &prefs->back);
-	add_key_binding(rw, "clicked", _("Skip forwards"), &prefs->forward);
-
 	gtk_signal_connect(GTK_OBJECT(ff), "event",
 			   GTK_SIGNAL_FUNC(skip_cb), GINT_TO_POINTER(1));
 	gtk_signal_connect(GTK_OBJECT(rw), "event",
 			   GTK_SIGNAL_FUNC(skip_cb), GINT_TO_POINTER(-1));
 
-/* Create goto button  */
+	/* Create goto button  */
 	gotobutton = gtk_button_new();
 	name = gnome_pixmap_file("tcd/goto.xpm");
 	pixmap = gnome_pixmap_new_from_file(name);
@@ -360,14 +485,25 @@ GtkWidget* create_buttons(void)
 	gtk_table_attach_defaults(GTK_TABLE(table), ff, 1, 2, 1, 2);
 	gtk_table_attach_defaults(GTK_TABLE(table), gotobutton, 2, 3, 1, 2);
 
+/* BOTTOM ROW */
 	b1 = make_button_stock(GNOME_STOCK_PIXMAP_PROPERTIES, edit_window, &prefs->tracked, _("Open track editor"));
-	b2 = make_button_stock(GNOME_STOCK_PIXMAP_PREFERENCES, preferences, NULL, _("Preferences"));
+	b2 = make_button_stock(GNOME_STOCK_PIXMAP_PREFERENCES, preferences, &prefs->preferences, _("Preferences"));
 	b3 = make_button_stock(GNOME_STOCK_PIXMAP_QUIT, quit_cb, &prefs->quit, _("Quit"));
 
 	gtk_table_attach_defaults(GTK_TABLE(table), b1, 0, 1, 2, 3);
 	gtk_table_attach_defaults(GTK_TABLE(table), b2, 1, 2, 2, 3);
 	gtk_table_attach_defaults(GTK_TABLE(table), b3, 2, 3, 2, 3);
 
+/* INVISIBLE BUTTONS */
+	b1 = make_button_stock(GNOME_STOCK_PIXMAP_VOLUME, mixer_cb, &prefs->mixer, _("Mixer"));
+	gtk_object_set_data(GTK_OBJECT(table), "mixer", b1);
+
+	add_invisible(table, _("Skip backwards"), "back",    'B', &prefs->back);
+	add_invisible(table, _("Skip forwards"),  "forward", 'F', &prefs->forward);
+	add_invisible(table, _("Fast Forward"),   "ff",      ']', &prefs->fast_forward);
+	add_invisible(table, _("Rewind"),         "rw",      '[', &prefs->rewind);
+	add_invisible(table, _("Volume Up"),      "up",      '>', &prefs->vol_up);
+	add_invisible(table, _("Volume Down"),    "down",    '<', &prefs->vol_down);
 	return table;
 }
 
@@ -499,6 +635,8 @@ void draw_titles(void)
 
 void draw_time_scanning(void)
 {
+	if (!tfont)
+		return;
 	gdk_gc_set_foreground( gc, &track_color );
 	gdk_draw_text(status_db,tfont,gc,4,39+(tfont->ascent+tfont->descent)-2, 
 		      _("(Scanning)"),
@@ -673,7 +811,15 @@ gint volume_changed( GtkWidget *widget, gpointer *data )
 {
 	if(!data)
 	{
-		tcd_set_volume(&cd, (int)floor(GTK_ADJUSTMENT(vol)->value));
+	        gfloat value = GTK_ADJUSTMENT(vol)->value;
+
+		if(prefs->squared_volume) {
+		    tcd_set_volume(&cd, (int)rint(value * value / 256.0));
+		}
+		else {
+		    tcd_set_volume(&cd, (int)rint(value));
+		}
+
 	}
 	draw_status();
 	return 1;
@@ -692,6 +838,7 @@ void make_goto_menu()
 	char buf[TRK_NAME_LEN];
 	int i;
 	GtkWidget *item;
+	gboolean data_track;
 
 	if(gotomenu)
 		gtk_widget_destroy(gotomenu);
@@ -699,12 +846,17 @@ void make_goto_menu()
 
 	for(i=1; i <= cd.last_t; i++)
 	{
-		g_snprintf(buf, TRK_NAME_LEN, "%2d - %s", i, cd.trk[C(i)].name);
+		data_track = (cd.trk[C(i)].toc.cdte_ctrl == CDROM_DATA_TRACK);
+
+		g_snprintf(buf, TRK_NAME_LEN, "%2d - %s", i,
+			   data_track ? "[Data]" : cd.trk[C(i)].name);
 		item = gtk_menu_item_new_with_label(buf);
 		gtk_widget_show(item);
 		gtk_menu_append(GTK_MENU(gotomenu), item);
-		gtk_signal_connect(GTK_OBJECT(item), "activate",
-				   GTK_SIGNAL_FUNC(goto_track_cb), GINT_TO_POINTER(i));
+		if (!data_track)
+			gtk_signal_connect(GTK_OBJECT(item), "activate",
+					   GTK_SIGNAL_FUNC(goto_track_cb),
+					   GINT_TO_POINTER(i));
 	}
 	if(gotoi) gtk_signal_disconnect(GTK_OBJECT(gotobutton), gotoi);
 	gotoi = gtk_signal_connect_object(GTK_OBJECT(gotobutton), "event",
@@ -719,6 +871,8 @@ void adjust_status_size(void)
 	int t;
 	max_title_width = 0;
 
+	if (!tfont)
+		return;
 	t = gdk_string_width(tfont, cd.artist);
 	if(max_title_width < t)
 		max_title_width = t;
@@ -771,7 +925,14 @@ static gint status_configure_event(GtkWidget *widget, GdkEventConfigure *event)
 		status_height = status_area->allocation.height;
 		first=FALSE;
 	}
-	GTK_ADJUSTMENT(vol)->value = (double)tcd_get_volume(&cd);
+
+	if(prefs->squared_volume) {
+	    GTK_ADJUSTMENT(vol)->value = sqrt((double)tcd_get_volume(&cd) * 256.0);
+	}
+	else {
+	    GTK_ADJUSTMENT(vol)->value = (double)tcd_get_volume(&cd);
+	}
+
 	gtk_signal_emit_by_name(GTK_OBJECT(vol),"value_changed", "no_update");
 	return TRUE;
 }
@@ -890,6 +1051,8 @@ void setup_time_display(GtkWidget *table)
 
 void setup_fonts(void)
 {
+	GtkWidget *dialog;
+
 	if(tfont)
 		gdk_font_unref(tfont);
 
@@ -898,6 +1061,40 @@ void setup_fonts(void)
 	else
 		tfont = NULL;
 
+	if (!tfont)
+	{
+		prefs->trackfont = g_strdup(DEFAULT_FONT);
+		tfont = gdk_fontset_load( prefs->trackfont );
+		tfont = 0;
+		if (tfont)
+		{
+			dialog = gnome_warning_dialog(
+				_("Cannot load requested font. Reloading default font."));
+			gtk_widget_show(dialog);
+		}
+		else
+		{
+			prefs->trackfont = g_strdup(EMERGENCY_FONT);
+			tfont = gdk_fontset_load( prefs->trackfont );
+			tfont = 0;
+			if (tfont)
+			{
+				dialog = gnome_warning_dialog(
+					_("Cannot load default Helvetica font. Loading any fixed width font."));
+				gtk_widget_show(dialog);
+			}
+			else
+			{
+				configured = FALSE; /* Don't update status in the background. */
+				dialog = gnome_error_dialog(
+					_("Cannot load fixed width font. Exiting."));
+				gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
+				gnome_dialog_run(GNOME_DIALOG(dialog));
+				exit(1);
+			}
+		}
+	}
+
 	if(sfont)
 		gdk_font_unref(sfont);
     
@@ -1004,6 +1201,85 @@ void setup_keys()
 	accel = gtk_accel_group_get_default();
 }    
 
+/*
+ * CORBA glue
+ */
+
+/*** App-specific servant structures ***/
+typedef struct {
+	POA_GNOME_GTcd servant;
+	PortableServer_POA poa;
+} impl_POA_GNOME_GTcd;
+
+/*** Implementation stub prototypes ***/
+static void impl_GNOME_GTcd__destroy(impl_POA_GNOME_GTcd * servant,
+				     CORBA_Environment * ev);
+static void
+impl_GNOME_GTcd_open_new_cd(impl_POA_GNOME_GTcd * servant,
+			    CORBA_Environment * ev);
+
+/*** epv structures ***/
+static PortableServer_ServantBase__epv impl_GNOME_GTcd_base_epv =
+{
+	NULL,                      /* _private data */
+	NULL,                      /* finalize routine */
+	NULL,                      /* default_POA routine */
+};
+static POA_GNOME_GTcd__epv impl_GNOME_GTcd_epv =
+{
+	NULL,                      /* _private */
+	(gpointer) & impl_GNOME_GTcd_open_new_cd,
+};
+ 
+/*** vepv structures ***/
+static POA_GNOME_GTcd__vepv impl_GNOME_GTcd_vepv =
+{
+	&impl_GNOME_GTcd_base_epv,
+	&impl_GNOME_GTcd_epv,
+};
+ 
+/*** Stub implementations ***/
+static GNOME_GTcd 
+impl_GNOME_GTcd__create(PortableServer_POA poa, CORBA_Environment * ev)
+{
+	GNOME_GTcd retval;
+	impl_POA_GNOME_GTcd *newservant;
+	PortableServer_ObjectId *objid;
+
+	newservant = g_new0(impl_POA_GNOME_GTcd, 1);
+	newservant->servant.vepv = &impl_GNOME_GTcd_vepv;
+	newservant->poa = poa;
+	POA_GNOME_GTcd__init((PortableServer_Servant) newservant, ev);
+	objid = PortableServer_POA_activate_object(poa, newservant, ev);
+	CORBA_free(objid);
+	retval = PortableServer_POA_servant_to_reference(poa, newservant, ev);
+	
+	return retval;
+}
+
+static void
+impl_GNOME_GTcd__destroy(impl_POA_GNOME_GTcd * servant, CORBA_Environment * ev)
+{
+	PortableServer_ObjectId *objid;
+	objid = PortableServer_POA_servant_to_id(servant->poa, servant, ev);
+	PortableServer_POA_deactivate_object(servant->poa, objid, ev);
+	CORBA_free(objid);
+
+	POA_GNOME_GTcd__fini((PortableServer_Servant) servant, ev);
+	g_free(servant);
+}
+
+static void
+impl_GNOME_GTcd_open_new_cd(impl_POA_GNOME_GTcd * servant,
+			    CORBA_Environment * ev)
+{
+	tcd_playtracks(&cd, cd.first_t, cd.last_t, prefs->only_use_trkind);
+}
+
+/*
+ * Main
+ */
+
 static char *CD_device = NULL;
 poptContext ctx;
 
@@ -1021,11 +1297,36 @@ int main (int argc, char *argv[])
 {
 	GtkWidget *table;
 	GdkColor black;
+	CORBA_ORB orb;
+	CORBA_Environment ev;
+	PortableServer_POA poa;
+	GNOME_GTcd gtcd_server;
+	struct sigaction sig_info;
 
 	bindtextdomain(PACKAGE, GNOMELOCALEDIR);
 	textdomain(PACKAGE);
 
-	gnome_init_with_popt_table("gtcd", VERSION, argc, argv, gtcd_popt_options, 0, &ctx);
+	CORBA_exception_init(&ev);
+	orb = gnome_CORBA_init_with_popt_table("gtcd", VERSION, &argc, argv, 
+					       gtcd_popt_options, 0, &ctx, 
+					       GNORBA_INIT_SERVER_FUNC, &ev);
+ 
+	/* Check for existing instance */
+	gtcd_server = goad_server_activate_with_id (NULL,
+						    "GOAD:gtcd:19990918",
+						    GOAD_ACTIVATE_EXISTING_ONLY,
+						    NULL);
+	
+	
+	if (gtcd_server != CORBA_OBJECT_NIL) {
+		GNOME_GTcd_open_new_cd(gtcd_server, &ev);
+		exit(0);
+	}
+
+	poa = (PortableServer_POA)CORBA_ORB_resolve_initial_references(orb, "RootPOA", &ev);
+	goad_server_register(CORBA_OBJECT_NIL,
+			     impl_GNOME_GTcd__create(poa, &ev),
+			     "GOAD:gtcd:19990918", "object", &ev);
 
 	cd.play_method = NORMAL;
 
@@ -1059,8 +1360,8 @@ int main (int argc, char *argv[])
 	/* Initialize some timers */
 	if(cd.isplayable) tcd_gettime(&cd);
     
-	gtk_timeout_add(1000, (GtkFunction)slow_timer, NULL);
-	gtk_timeout_add(500, (GtkFunction)fast_timer, NULL);
+	slow_timeout = gtk_timeout_add(1000, (GtkFunction)slow_timer, NULL);
+	fast_timeout = gtk_timeout_add(500, (GtkFunction)fast_timer, NULL);
 	titlelabel_f = TRUE;
 	
 	gnome_app_set_contents(GNOME_APP(window), main_box);
@@ -1070,7 +1371,12 @@ int main (int argc, char *argv[])
 	gdk_color_black( gtk_widget_get_colormap(status_area), &black);
 	gdk_window_set_background( status_area->window, &black);
     
-	signal(SIGUSR2, reload_info);
+	memset (&sig_info, 0, sizeof(sig_info));
+	sig_info.sa_handler = reload_info;
+	sigaction(SIGUSR2, &sig_info, NULL);
+
+	PortableServer_POAManager_activate
+		(PortableServer_POA__get_the_POAManager(poa, &ev), &ev);
     
 	gtk_main();
 	return 0;
